package org.freecompany.imp;

import org.freecompany.brimstone.module.Module;
import org.freecompany.imp.index.Generator;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.util.Arrays;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.PropertyHelper;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.PatternSet;
import org.apache.tools.ant.types.FileSet;

import static org.apache.tools.ant.Project.MSG_DEBUG;
import static org.freecompany.imp.Constants.FILESET_COMPILE;
import static org.freecompany.imp.Constants.FILESET_RUN;
import static org.freecompany.imp.Constants.FILESET_PACKAGE;
import static org.freecompany.imp.Constants.FILESET_TEST;
import static org.freecompany.imp.Constants.PATH_COMPILE;
import static org.freecompany.imp.Constants.PATH_RUN;
import static org.freecompany.imp.Constants.PATH_TEST;
import static org.freecompany.imp.Constants.PROPERTY_GROUP;
import static org.freecompany.imp.Constants.PROPERTY_NAME;
import static org.freecompany.imp.Constants.PROPERTY_SKIP;
import static org.freecompany.imp.Constants.PROPERTY_VERSION;
import static org.freecompany.imp.Constants.REFERENCE_INDEX;
import static org.freecompany.imp.Constants.REFERENCE_MODULE;
import static org.freecompany.imp.Constants.REFERENCE_SUBPROJECTS;
import static org.freecompany.imp.Constants.VERSION_SNAPSHOT;
import static org.freecompany.imp.Util.getCompile;
import static org.freecompany.imp.Util.getRun;
import static org.freecompany.imp.Util.getPackage;
import static org.freecompany.imp.Util.getProperties;
import static org.freecompany.imp.Util.getTest;
import static org.freecompany.imp.Util.initialize;
import static org.freecompany.imp.Util.loadProperties;

/**
 * Models the index structure for an individual project.
 */
public class Index extends Task {

	protected String version;
	protected List< Dependency> dependencies = new LinkedList< Dependency>();
	protected List< Repository> repositories = new LinkedList< Repository>();
	protected List< AntOverride> subprojects = new LinkedList< AntOverride>();
	protected String name;
	protected String group;
	protected List< String> types;

	protected String format = "antlib:org.freecompany.imp:magic";
	protected File src;

	public void setProject( final Project project) {
		this.name = project.getName();
		super.setProject( project);
	}

	protected void include( final String name) {
		final Import include = new Import();
		include.setUri( name);
		include.setOwningTarget( getOwningTarget());
		include.setLocation( getLocation());
		include.setProject( getProject());
		include.execute();
	}

	/**
	 * Configures this index structure using the values and objects set on this index during
	 * the build initialization.  If the 'src' attribute is set the file is parsed and processed
	 * using the interpreter specified in the 'type' attribute.  Next any subprojects of this index,
	 * whether they were directly nested in the build file or specified in the external index, are
	 * executed.  Finally, the required properties for this build are set so that the local target
	 * is ready to execute.
	 * </p>
	 * Type targets are handled as a special case.  For each type listed in the index element's <code>type</code>
	 * attribute, a check is made to see if a target for that type is already regsitered in the project.  If not,
	 * then the index code will automatically attempt to resolve a template corresponding to the type against the
	 * base module.  For example, providing the index type <code>war</code> will cause the index to search for
	 * any locally definined target named <code>war</code> or one in any imports in the build file.  If none is
	 * found the index will attempt to import the uri <code>#war.template</code>.  This fragment is resolved against
	 * the core module which, if the template is provided with Imp, will maintain an alias mapping that common name to
	 * the specific template version and URI.
	 * <p/>
	 * After setting up the complete environment for building, this method will finally iterate through the
	 * Imp specific property types to execute them.
	 */
	public void execute() throws BuildException {
		initialize( getProject());

		try {
			final String home = System.getProperty( "user.home") + File.separator + ".imp" + File.separator + group;
			loadProperties( getProject(), "project.properties", true);
			loadProperties( getProject(), home + File.separator + name + ".properties", true);
			loadProperties( getProject(), home + ".properties", true);
		} catch ( IOException e) {
			throw new BuildException( e);
		}
		
		include( "#builder.template");

		final Hashtable targets = getProject().getTargets();
		try {
			if ( types != null) {
				log( "Importing types '" + types + "'.", MSG_DEBUG);
				for ( String type : types) {
					log( "Loading internal template '" + type + "'.", MSG_DEBUG);
					final String fragment = type + ".template";
					final URL url = (( Module) getProject().getReference( REFERENCE_MODULE)).getComponent( fragment);
					if ( url != null) {
						log( "Found reference at '" + url + "'.", MSG_DEBUG);
						include( url.toString());
					} else if ( !targets.containsKey( type)) throw new BuildException( "Could not find type '" + type + "' definition.");
				}
			}
		} catch ( IOException e) {
			throw new BuildException( e);
		}

		setupRepositories();
		setupIndexFile();
		setupDependencies();
		setupReferences();
		
		try {
			Util.prepareReplacements( getProject());
		} catch ( IOException e) {
			throw new BuildException( e);
		}

		if ( types != null) {
			for ( String type : types) {
				String properties = type + ".properties";
				if ( targets.containsKey( properties)) {
					log( "Setting properties for type '" + type + "' by calling '" + properties + "'.", MSG_DEBUG);
					getProject().executeTarget( properties);
				}
			}
		}

		log( "Executing deferred properties.", MSG_DEBUG);
		final PropertyHelper helper = PropertyHelper.getPropertyHelper( getProject());
		for ( PropertyTask property : getProperties( getProject())) {
			property.setValue( helper.replaceProperties( null, property.getValue(), null));
			property.executeNow();
		}
	}

	/**
	 * Sets up any locally defined repositories in the index element.
	 */
	protected void setupRepositories() {
		for ( Repository repository : repositories) {
			repository.setProject( getProject());
			repository.execute();
		}
	}

	/**
	 * If this index specifies a source document with the <code>src</code> attribute it will be parsed
	 * according to the format specified.  The information gathered from parsing the index file is added
	 * to that which is specified directly in the index element, except where it is contradictory, in which
	 * case the index file takes presidence.
	 */
	protected void setupIndexFile() {
		getProject().addReference( REFERENCE_SUBPROJECTS, subprojects);
		if ( src != null) {
			try {
				Generator generator = ( Generator) getProject().createDataType( format);
				if ( generator == null) throw new IllegalStateException();
				generator.setProject( getProject());
				generator.setIndex( this);
				generator.setName( name);
				generator.setSource( new FileInputStream( src));
				generator.execute();
			} catch ( IllegalStateException e) {
				throw new BuildException( "The index type attribute was not a recognized index file format type.  Allowed type(s) '" + getAllowedIndexTypes() + "'.");
			} catch ( ClassCastException e) {
				throw new BuildException( "The index type attribute '" + format + "' is not a valid index file format data type.  Allowed type(s) '" + getAllowedIndexTypes() + "'.");
			} catch ( Exception e) {
				throw new BuildException( e);
			}
		}
	}

	protected void setupDependencies() {
		for ( Dependency dependency : dependencies) dependency.execute();
	}

	/**
	 * Takes the local object sets that have been built up by defining dependencies, resolution tasks
	 * and index file parsing and publishes them to the build project context.  These references will
	 * be available to both internal Ant classes as well as targets in the build script.
	 */
	protected void setupReferences() {
		getProject().setProperty( PROPERTY_NAME, name);
		getProject().setProperty( PROPERTY_VERSION, VERSION_SNAPSHOT);
		if ( group == null) getProject().setProperty( PROPERTY_SKIP, Boolean.TRUE.toString());
		else {
			try {
				buildPath( PATH_COMPILE, getCompile( getProject()));
				buildPath( PATH_TEST, getTest( getProject()));
				buildPath( PATH_RUN, getRun( getProject()));
				buildFileset( FILESET_COMPILE, getCompile( getProject()));
				buildFileset( FILESET_TEST, getTest( getProject()));
				buildFileset( FILESET_RUN, getRun( getProject()));
				buildFileset( FILESET_PACKAGE, getPackage( getProject()));
			} catch ( Exception e) {
				throw new BuildException( e);
			}
		}
		getProject().addReference( REFERENCE_INDEX, this);
	}

	/**
	 * For the given list of dependencies ths method builds a corresponding Ant path element which is
	 * inserted by reference into the project.  This allows targets and tasks in the scripts to reference
	 * them without requiring Java code.
	 */
	protected void buildPath( final String name, final Collection< Dependency> list) throws Exception {
		Path path = new Path( getProject());
		for ( Dependency dependency : list) {
			Path.PathElement element = path.createPathElement();
			element.setLocation( dependency.getFile());
		}
		getProject().addReference( name, path);
	}

	/**
	 * This is an inoptimal solution to building filesets for use in the build system.  Because
	 * this method assumes that all dependencies share a common filesystem root it would certainly
	 * break if this were not the case.  This is not an issue on any Unix system as they have only
	 * one root, nor on Windows systems with the current URI handlers we have available.  But be
	 * warned...
	 */
	 protected void buildFileset( String name, Collection< Dependency> list) throws Exception {
		File[] roots = File.listRoots();
		File root = null;

		FileSet fileset = new FileSet();
		for ( Dependency dependency : list) {
			if ( !dependency.isModule()) {
				File file = dependency.getFile();
				String path = file.getCanonicalPath();
				String current = null;
				for ( File tmp : roots) {
					current = tmp.getCanonicalPath();
					if ( path.startsWith( current)) {
						root = tmp;
						break;
					}
				}
				PatternSet.NameEntry entry = fileset.createInclude();
				entry.setName( file.getCanonicalPath().substring( current.length()));
			}
		}
		fileset.setDir( root);
		getProject().addReference( name, fileset);
	}

	/**
	 * Generates a comma separated list of possible index types.  This is generated by examining all registered
	 * types for compatibility.  This list is only really useful for debugging purposes, though it could also
	 * potentially be used for finding a default parser implementation to use instead of the hardcoded default
	 * we currently use.
	 */
	@SuppressWarnings({ "unchecked" })
	protected String getAllowedIndexTypes() {
		StringBuffer buffer = new StringBuffer();
		Hashtable< String, Class> map = getProject().getDataTypeDefinitions();
		for ( Map.Entry< String, Class> entry : map.entrySet()) {
			if ( Generator.class.isAssignableFrom( entry.getValue())) {
				if ( buffer.length() != 0) buffer.append( ",");
				buffer.append( entry.getKey());
			}
		}
		return buffer.toString();
	}

	public void setGroup( String group) {
		getProject().setProperty( PROPERTY_GROUP, group);
		this.group = group;
	}

	public void setTypes( String types) {
		this.types = Arrays.asList( types.split( ","));
	}

	public void setFormat( String format) {
		this.format = format;
	}

	public void setSrc( File src) {
		this.src = src;
	}

	public void addAnt( AntOverride ant) {
		subprojects.add( ant);
	}

	public void addRepository( Repository repository) {
		repositories.add( repository);
	}

	public void addDependency( Dependency dependency) {
		dependencies.add( dependency);
	}

	public String getName() {
		return name;
	}

	public String getGroup() {
		return group;
	}

	public List< String> getTypes() {
		return types;
	}

	public List< AntOverride> getSubprojects() {
		return subprojects;
	}
}
